#include "pnc.h"

typedef unsigned char boolean;
#define False ((boolean) 0)
#define True  ((boolean) 1)

typedef struct {
	short count;
	boolean sense;
} central_barrier_t;

PNC_cta_param_block count_cta_parms;

short max_delay;
boolean local_sense;

extern central_barrier_t *shared_array;
extern int num_nodes;


void INIT()
{
	shared_array->sense = False;
	shared_array->count = num_nodes;
	local_sense = False;

#ifdef DELAY_MAXIMUM
	max_delay = DELAY_MAXIMUM * (num_nodes - 1);
#endif DELAY_MAXIMUM

	/* pre-initialize control block for all atomic-clear-then-adds */
	count_cta_parms.cta_mask = 0;
	count_cta_parms.cta_incr = -1;
	count_cta_parms.cta_addr = &shared_array->count;
}

void CALL()
{
	register central_barrier_t *bp = shared_array;

	local_sense ^= True;
	perform_cta = &count_cta_parms;
	if (count_cta_parms.cta_oldvalue == 1) {
		bp->count = num_nodes;
		bp->sense = local_sense;
	}
	else {
#ifdef BACKOFF
		register long next_delay = DELAY_CONSTANT;
#endif BACKOFF
		while(bp->sense != local_sense) {
#ifdef BACKOFF
			register long current_delay;
#ifdef DELAY_MAXIMUM
			if (next_delay > max_delay) current_delay = max_delay;
			else 
#endif DELAY_MAXIMUM
			{
				current_delay = next_delay;
#ifdef DELAY_BASE
				next_delay <<= DELAY_BASE;
#endif DELAY_BASE
			}
			while(current_delay--);
#endif BACKOFF
		}
	}
}
