#include "types.h" 
#include "pnc.h" 
#include "shared.h" 

typedef short boolean;
#define True ((boolean) 1)
#define False ((boolean) 0)

typedef struct {
	volatile boolean my_flag;
	volatile boolean *right;
	volatile boolean *left;
	boolean dummy;
} treenode;


extern treenode *shared_array[];
extern int num_nodes;
extern int vpid;

PNC_cta_param_block cta_block;

treenode *mynode;
short barriercount;


void central_treewakeup_barrier_init()
{
	cta_block.cta_mask = 0;
	cta_block.cta_incr = 1;

	/* centralized counter */
	cta_block.cta_addr = (short *) &CD->tbarrier_count;

	mynode = shared_array[vpid];

	if (vpid == 0) {
		barriercount = num_nodes - 1;
		/* initialize central counter */
		CD->tbarrier_count = 0;
	}

	{
		int child_id = vpid * 2 + 1;

		if (child_id < num_nodes) 
			mynode->left = &shared_array[child_id++]->my_flag;
		else mynode->left = &mynode->dummy;

		if (child_id < num_nodes) 
			mynode->right = &shared_array[child_id]->my_flag;
		else mynode->right = &mynode->dummy;
	}
}


void central_treewakeup_barrier()
{
	register treenode *mynode_reg = mynode;

	if (vpid) { /* interior or leaf node in wakeup tree */
		mynode_reg->my_flag = True;
		perform_cta = &cta_block;
		while(mynode_reg->my_flag);
	}
	else { /* root node in wakeup tree */
		/* node 0 (root) waits for all arrivals */
		while (CD->tbarrier_count != barriercount); 

		/* now everybody is up to the barrier */
		CD->tbarrier_count = 0;
	}
	*mynode_reg->right = False;
	*mynode_reg->left = False;
}
