#include <public.h>
#include <stdio.h>

#include "shared.h"

#define STACK_SIZE (60 K)
#define NUM_ARGS 6
#define MAX_NODES SWITCH_SIZE
#define TICKS_PER_SEC 16000
#define MY_NODE -1

int num_nodes = 0;
int node_list[MAX_NODES];
EH children[MAX_NODES];
char *process_argv[NUM_ARGS];/* Pointers to arguments to the process */
OID shared_obj_id;
int interrupts = 0;
int remote = 0;

centralized_data_t *CD;

void build_proc_array(int id, long nodes)
{
	if (id == pnn) return;
	if (num_nodes == nodes) return;
	node_list[num_nodes++] = id;
}

main(int argc, char **argv)
{
	void spawn_children();

	void For_All_Nodes(void *, long);
	long nodes = 2;
	remote = 0;

	while (--argc > 0 && (*++argv)[0] == '-') {
	    switch (*(argv[0]+1)) {

		case 'r' :
		    remote = 1;
		    break;

		case 'i' :
		    interrupts = 1;
		    break;

		case 'p' :
		    nodes = atoi(argv[0]+2);
		    break;
		
		default:
		    printf("Usage: ");
		    printf("main [-r] [-i] [-p<nodes>] \n");
		    exit(1);
	    } /* end switch */
	} /* end while more arguments */

	For_All_Nodes((void *)build_proc_array, nodes);
	shared_obj_id = Make_Obj(0,MY_NODE,sizeof(centralized_data_t), RW_rw_);
	CD = (centralized_data_t *) Map_Obj(shared_obj_id,0,RW_rw_);
	CD->num_nodes = num_nodes;
	CD->barrier_flag = 0;
	CD->barrier_count = 0;
	spawn_children();
	child_wait();
}

child_wait()
{
	int proc_cnt;
	for (proc_cnt = 0; proc_cnt < num_nodes; proc_cnt++) {
		MWait(children[proc_cnt],0);
		Del_Obj(children[proc_cnt]);
	}
	Unmap_Obj(shared_obj_id,CD);
	Del_Obj(shared_obj_id);
}

#define MAX_ARG_SIZE 20     /* Max string argument size (in bytes) */
char string_space[200];     /* Stores argument strings to child process */

void spawn_children()
{
    OID template_OID;
    OID arg_OID;
    CP_ARGS process;	/* Record to hold arguments to Make_Process */
    int proc_cnt;
    int arg_cnt;

    catch

	/* Allocate string space for each argument */
	for (arg_cnt = 0; arg_cnt < NUM_ARGS; arg_cnt++) {
	    process_argv[arg_cnt] = &string_space[arg_cnt*MAX_ARG_SIZE];
	}

	/* Store values in argument strings */

	process_argv[0] = "spintest";
	/* Argument [1] is the vpid and is set individually below */
	sprintf(process_argv[2],"%d",shared_obj_id);
	sprintf(process_argv[3],"%d",interrupts);
	sprintf(process_argv[4],"%d",remote);
	/* Argument [5] is the vpid and is set individually below */

	catch
		template_OID = Find_Value("spintest.68", NTYPE_OBJ);
	onthrow
		when(TRUE) {
			template_OID = Load_Process("spintest.68",-1,200,STACK_SIZE);
		}
	endcatch;

	/* Only set fields not set in template or in Fork_Process */
	/* process.cp_pn supplied below */
	process.cp_type = CP_CHILD;     /* Make the new process a child */
	process.cp_flags = CP_PRIV;     /* run on creation */
	process.cp_name = NULL;         /* No template name provided here */
	process.cp_spec = template_OID; /* Template OID for process */
	process.cp_prot = R_Xr_x;
	process.cp_prio = 1;
	process.cp_slice = 10;
	process.cp_need = 20;
	process.cp_stksz = STACK_SIZE;
	process.cp_stkloc = 0;
	/* process.cp_arg1 supplied below */
	process.cp_arg2 = NULL;
	process.cp_start = 0;
	process.cp_segs = 100;
	process.cp_msegsize = 0;
	process.cp_exit_eh = NULL;


	for (proc_cnt = 0; proc_cnt < num_nodes; proc_cnt++) {

	    /* Fork another process */


	    /* Pass the process its virtual processor number */
	    sprintf(process_argv[1],"%d",proc_cnt);

		children[proc_cnt] = Make_Event(0,0,0,0);
	    /* Pass the process an exit event */
		sprintf(process_argv[5],"%d",children[proc_cnt]);

	    /* Create arguments for the process */
	    arg_OID = Create_Args(node_list[proc_cnt],NUM_ARGS,process_argv,NULL);

	    process.cp_pn = node_list[proc_cnt];	/* Node to run process on */
	    process.cp_arg1 = arg_OID; 	/* Argument object */

	    /* Create the process */
		{ 
			OID poid = Make_Process(&process);

			/* Give the argument object to the child */
			Disown_Obj(arg_OID, poid);
		}
	} /* end for */

    onthrow
	when (TRUE) {
	    printf("* Error in %s procedure spawn_children *\n",
			"spintest_master");
	    printf("%s\n",throwtext);
	    rethrow;
	}
    endcatch;

}  /* spawn_children */
