#include <public.h>
#include <stdio.h>

#define MY_NODE -1

#include "shared.h"

centralized_data_t *CD;

#define ONE_K_SEGMENT 1000
char *shared_array[256];

EH exit_eh;
OID shared_obj_id;
int vpid;
int num_nodes;

extern void INIT();
extern void CALL();

main(argc,argv)
int argc;
char **argv;
{
	int i;
	int interrupts;
	int remote;

	vpid = atoi(argv[1]);
	shared_obj_id = atoi(argv[2]);
	interrupts = atoi(argv[3]);
	remote = atoi(argv[4]);
	exit_eh = atoi(argv[5]);
	Set_Process_Exit_Event(CurPCB->p_oid,exit_eh);

	CD = (centralized_data_t *) Map_Obj(shared_obj_id,0,RW_rw_);
	CD->shared_segoid[vpid] = Make_Obj(0,MY_NODE,ONE_K_SEGMENT, 0);
	num_nodes = CD->num_nodes;
	simple_barrier();

	/*	before the barrier, each processor has entered the id for its shared 
		segment into the table CD->shared_segoid.  map in shared area for 
		each processor using the segment ids from the global table */
	for(i=0;i<num_nodes;i++) {
		shared_array[i] = (char *) 
			Map_Obj(CD->shared_segoid[i],0,RW_rw_);
	}

	if (remote) {
		int segment = ((((unsigned int) shared_array[vpid]) >> 16) & 0xff);
        short kmode = entkerin; /* enter kernel mode */
        usrsar(segment) |= 0xc0; /* change memory subspace to remote */
        AsAr = kmode; /* exit kernel mode */
	}

	/*	the shared areas are set up, initialize pointers in the 
		shared area that are used for tight barrier synchronization */

	if (vpid == 0) {
		printf("%s",BARRIERSTRING);
		if (interrupts) printf("-int");
		else printf("-noint");
		if (remote) printf("-remote"); 
		printf("\n");
	}

	/* local initialization of tight barrier */
	INIT();

	simple_barrier();

	/* 	the tight barrier initialization is complete, synchronize
		before performing the timing tests */


#define BARRIER_ITERATIONS 100000
	{
		register int i;
		long start,end;
		short kmode;
		if (!interrupts) kmode = entkerin; /* inhibit interrupts */

		CALL();

		/*	perform the timing tests */

		start = rtc;
		for(i=BARRIER_ITERATIONS;i != 0; i--) CALL();
		end = rtc;

		if (!interrupts) AsAr = kmode; /* enable interrupts */

		if (vpid == 0) {
		printf("Barrier test %d processes, %d iterations, time = %d ticks\n",
			num_nodes, BARRIER_ITERATIONS, end - start);
		}
	}


	/*	terminate the computation signalling the main program that forked
		all of the slaves for the spintest */

	for(i=0;i<num_nodes;i++) Unmap_Obj(CD->shared_segoid[i],shared_array[i]);
	simple_barrier();
	Del_Obj(CD->shared_segoid[vpid]);
	Unmap_Obj(shared_obj_id,CD);
}
