#define NULL	0
#define P 		128
#define LogP	7	
#define private
#define shared 

typedef unsigned char boolean;
#define False ((boolean) 0)
#define True ((boolean) 1)

typedef unsigned char role_enum;
#define LOSER ((role_enum) 1)
#define WINNER ((role_enum) 2)
#define CHAMPION ((role_enum) 4)
#define NOOP ((role_enum) 8)
#define DROPOUT ((role_enum) 16)

typedef struct {
	volatile boolean *opponent;
	role_enum role;
	volatile boolean flag;
} round_t;

extern shared round_t *shared_array[];

private boolean sense = True;
private round_t *myrounds;

extern private int vpid;
extern shared int num_nodes;

void tournament_barrier_init()
{
	int k, log_num_nodes = 0;
	myrounds = &shared_array[vpid][0];
	for(k = 1; k < num_nodes; log_num_nodes++, k <<= 1); 
	for(k = 0; k <= LogP; k++) myrounds[k].role = DROPOUT; /* default */
	for(k = 0; k < log_num_nodes; k++) {
		int round = k + 1;
		int vpid_mod_2_sup_k_plus_1 = vpid % (1 << (k+1));
		if (vpid_mod_2_sup_k_plus_1 == 0) {
			int partner = vpid + (1 << k);
			if (partner < num_nodes) {
				myrounds[round].role = WINNER;
				myrounds[round].opponent = 
					&((shared_array[partner])[round].flag);
			} else {
				myrounds[round].role = NOOP;
				myrounds[round].opponent = (boolean *) NULL;
			}
		}
		else if (vpid_mod_2_sup_k_plus_1 == (1 << k)) {
			myrounds[round].role = LOSER;
			myrounds[round].opponent = 
				&((shared_array[vpid - (1<<k)])[round].flag);
			break;
		}
	}
	if (vpid == 0) myrounds[k].role = CHAMPION;
	for(k = 0; k < LogP; k++) myrounds[k].flag = False; 
}

void tournament_barrier()
{
	register round_t *round = myrounds;
	for(;;) {
		if ((++round)->role & LOSER) {
			*(round->opponent) = sense;
			while (round->flag != sense);
			goto release;
		} else {
			if (round->role & WINNER) while (round->flag != sense);
			else if (round->role & CHAMPION) {
				while (round->flag != sense);
				*(round->opponent) = sense;
				goto release;
			}
		}
	}
release:
	for(;;) {
		if ((--round)->role & WINNER) *(round->opponent) = sense;
		else if (round->role & DROPOUT) break;
	}
	sense ^= True;
}
