typedef unsigned char boolean;
#define True ((boolean) 1)
#define False ((boolean) 0)

typedef union {
	long whole;
	boolean parts[4];
} whole_and_parts;

typedef struct {
	volatile boolean parentsense;
	volatile boolean *parentflag;
	volatile boolean *childwait[2];
	whole_and_parts havechild;
	volatile whole_and_parts childnotready;
	boolean dummy;
} treenode;

extern treenode *shared_array[];
extern int num_nodes;
extern int vpid;

boolean sense;
treenode *mynode;

void tree_barrier_init()
{
	int i,child_id;

	mynode = shared_array[vpid];
	if (vpid == 0) mynode->parentflag = &mynode->dummy;
	else {
		int parentid = (vpid -1)/4;
		int my_index = vpid - (parentid * 4) - 1;

		mynode->parentflag = 
			&shared_array[parentid]->childnotready.parts[my_index];
	}
	for(i=0,child_id = 2 * vpid + 1;i<2;i++,child_id++) {
		if (num_nodes > child_id)  /* have child i in wakeup tree */
			mynode->childwait[i] = &shared_array[child_id]->parentsense;
		else  /* don't have child i in wakeup tree */
			mynode->childwait[i] = &mynode->dummy;
	}

	for(i=0,child_id = 4 * vpid + 1;i<4;i++,child_id++) {
		 /* have child i in arrival tree */
		mynode->havechild.parts[i] = (num_nodes > child_id);
	}

	mynode->childnotready.whole = mynode->havechild.whole;
	mynode->parentsense = True;
	sense = False;
}


void tree_barrier()
{
	register treenode *mynode_reg = mynode;

	while (mynode_reg->childnotready.whole);
	mynode_reg->childnotready.whole = mynode_reg->havechild.whole; 
	*mynode_reg->parentflag = False;
	{
		register boolean sense_reg = sense;
		if (vpid) while(mynode_reg->parentsense != sense_reg);
		*mynode_reg->childwait[0] = sense_reg;
		*mynode_reg->childwait[1] = sense_reg;
	}
	sense ^= True;
}
