/***************************************************************** 
	from the expanded TR version of the ICPP paper
	guarantees FIFO ordering
	spins locally if you have coherent caches
	avoids the flurry of bus traffic when a lock is released
	requires space for the lock linear in the number of processors
	may have a counter overflow problem (fixed here)
 *****************************************************************/

#include <public.h>
#include "util.h"
#include "simple.h"
#include "ticket.h"
#include "anderson.h"
#include "mcs.h"
#include "shared.h"


extern short vpid;
extern short num_nodes;
short _num_nodes;
short too_big;
short too_big_neg;

#ifndef USE_ASM
	fail
#endif USE_ASM

#define A(s) asm volatile(s);

static struct PNC_am anderson_cta_block;
void init_A()
{
	too_big = ((1<<14) / num_nodes)* num_nodes;
	too_big_neg = -too_big;
	_num_nodes = - num_nodes;

	if (vpid == 0) {
		shared[vpid]->distributed_locks.a_link.locked_flag = FALSE;
		SV->a.queue_last = 0;
	}
	else shared[vpid]->distributed_locks.a_link.locked_flag = TRUE;

	anderson_cta_block.am_mask = 0;
	anderson_cta_block.am_incr = 1;
}


static void foo()
{
A ("	| assembly version of anderson lock 								")
A ("	| parameter: pointer to lock in a1									")
A ("	.text																")
A ("	.even																")
A (".globl anderson_lock													")
A ("anderson_lock:															")
A ("	| don't need an fp													")
A ("	| no non-volatiles to save											")
A ("	movl #_anderson_cta_block,a0										")
A ("	| fill in control block for atomic increment 						")
A ("	movl a1,a0@				| fill in &alock.queue_last in ctl blk		")
A ("	movl a0,0xffffd100		| fetch-and-phi --> my_place				")
A ("	movw a0@(8),d0			| put my_place in d0						")
A ("	cmpw _num_nodes,d0		| compare my_place and num_nodes       		")
A ("	blt compute_queue_slot	| if (my_place >= num_nodes)                ")
A ("	subw _num_nodes,d0		| my_place -= num_nodes                     ")
A ("	bgt compute_queue_slot  | if (my_place == 0) 						")
A ("	movw __num_nodes,a0@(6) | _anderson_cta_block.incr = -num_nodes		")
A ("	movl a0,0xffffd100		| prevent overflow of queue_last			")
A ("	movw #1,a0@(6)			| restore _anderson_cta_block.incr = 1		")
A ("compute_queue_slot:			| invariant: 0 <= my_place < num_nodes      ")
A ("	movw d0,d1    			| put my_place in d1						")
A ("   	| compute address of queue slot: shared[my_place % num_nodes]		")
A ("    lea _shared,a0			| array of pointers to distributed slots	")
A ("	aslw #2,d1				| compute array offset from slot index		")
A ("	movl a0@(0,d1:w),a0		| grab address of my queue slot				")
A ("    tstw a0@				| is queue slot locked? 					")
A ("	bne do_poll				| continue polling							")
A ("    movw #1,a0@				| lock queue slot for next time through		")
A ("	rts																	")
A ("do_poll:																")
A ("	movl a7@+,a1			| pop return address (setup fast return)	")
A ("poll_again:																")
A ("    tstw a0@				| is queue slot locked? 					")
A ("	bne poll_again			| if not, poll again						")
A ("    movw #1,a0@				| lock queue slot for next time through		")
A ("	jmp a1@					| return to caller 							")
}
