/***************************************************************** 
	John and my algorithm
	guarantees FIFO ordering
	spins locally even if you don't have coherent caches
	no unusual space requirements
 *****************************************************************/

#include <public.h>
#include "util.h"
#include "simple.h"
#include "ticket.h"
#include "anderson.h"
#include "mcs.h"
#include "shared.h"

#define fetch_and_store_16(addr,val) Atomic_cTa(addr,0xffff,val)

/*
	We don't have a 32-bit fetch-and-store, so we can't use pointers
	(or paddrs) for the algorithm.  What we do instead is store the
	node number and keep track of the locations (virtual and physical)
	of the unique link record for each node.
*/

#define TRUE 1
#define FALSE 0

extern short vpid;
mlink *my_link; 		/* location of my own link record */

#ifndef USE_ASM
	fail
#endif USE_ASM

struct PNC_am mcs_cta_block;
struct PNC_am mcs_cta_ublock;

void init_M ()
{
	/* global initialization */
	if (vpid == 0) SV->m.tail = NONE;

	/* local initialization */
	my_link = &shared[vpid]->distributed_locks.m_link;

	/* hand-coded assembler uses its own special ctl blocks for speed */
	mcs_cta_block.am_mask = 0xffff;
	mcs_cta_block.am_incr = vpid;

	mcs_cta_ublock.am_mask = 0xffff;
	mcs_cta_ublock.am_incr = NONE;
}

#define A(s) asm volatile (s);

static void foo()
{
A ("	| assembly version of mellor-crummey/scott lock						")
A ("	| parameter: pointer to lock in d0									")
A ("	| parameter: return address in d1 									")
A ("	.text																")
A ("	.even																")
A (".globl mlock_lock														")
A ("mlock_lock:																")
A ("	| no non-volatiles to save											")
A ("	| no need for an fp													")
A ("	movl _my_link,a1 													")
A ("	clrl a1@				| my_link->successor = nil					")
A ("	| unsigned short oldtail = fetch_and_store_16 (&tail, vpid) 		")
A ("	movl #_mcs_cta_block,a0												")
A ("	movl d0,a0@				| msc_cta_block.addr = &lockaddr->tail		")
A ("	movl a0,0xffffd100		| execute fetch-and-store					")
A ("	movw a0@(8),d0			| oldtail									")
A ("	bpl local_spin			| NONE is the only negative value			")
A ("	movl d1,a0				| return address from d1 (setup return)		")
A ("	jmp a0@					| return									")
A ("local_spin:																")
A ("	movw #1,a1@(4)			| my_link->locked = TRUE					")
A ("	|	shared[oldtail]->distributed_locks.mlink.successor = my_link 	")
A ("	lea _shared,a0														")
A ("	aslw #2,d0															")
A ("	movl a0@(0,d0:w),a0													")
A ("	movl a1,a0@															")
A ("do_poll:																")
A ("	movl d1,a0				| return address from d1 (setup return)		")
A ("	lea a1@(4),a1			| simplify test								")
A ("poll_again:																")
A ("	tstw a1@				| has my locked bit been unset				")
A ("	bne poll_again			| if not, continue polling					")
A ("	jmp a0@					| return									")

A ("	.even																")
A (".globl mlock_unlock_helper 												")
A ("mlock_unlock_helper:													")
A ("	| no non-volatiles to save											")
A ("	| no need for an fp													")
A ("	| parameters: my_link in a1, lockaddr in d0							")
A ("no_successor:                     	                                    ")
A ("	| tail = fetch_and_store_16 (&tail, NONE)	 						")
A ("	| INVARIANT:														")
A ("	| _mcs_cta_ublock mask 0xffff and increment NONE					")
A ("	movl #_mcs_cta_ublock,a0											")
A ("	movl d0,a0@					| mcs_cta_ublock.addr = &this->tail		")
A ("	movl a0,0xffffd100			| execute								")
A ("	movw a0@(8),d1				| oldtail								")
A ("	cmpw _vpid,d1				| was I still the tail?					")
A ("	bne queue_non_empty													")
A ("	rts																	")
A ("queue_non_empty:														")
A ("	movw d1,a0@(6)				| mcs_cta_ublock.incr = oldtail			")
A ("	movl a0,0xffffd100			| execute								")
A ("	movw #0xffff,a0@(6)			| mcs_cta_ublock.incr = NONE			")
A ("	movw a0@(8),d1				| usurper								")
A ("	bpl usurper_found													")
A ("await_successor:														")
A ("    movl a1@,d0															")
A ("    beq await_successor													")
A ("    movl d0,a1					| head = my_link->successor;			")
A ("	clrw a1@(4)					| head->locked = FALSE					")
A ("	rts																	")
A ("usurper_found:															")
A ("    |   shared[usurper]->local_link.successor = head 					")
A ("    lea _shared,a0                                                 		")
A ("	aslw #2,d1															")
A ("	movl a0@(0,d1:w),a0													")
A ("await_successor2:														")
A ("    movl a1@,d0															")
A ("    beq await_successor													")
A ("    movl d0,a0@															")
A ("	rts																	")

}
