/*
	spin lock alternatives for testing purposes
*/

#include <public.h>
#include "mcs.h"
#include "simple.h"
#include "ticket.h"
#include "anderson.h"
#include "shared.h"
#include "barrier.h"
#include "util.h"
#include "synchtest.h"


short nodes_in_test;
ticks start_time, end_time; /* for use in following macro */
long repeat_count;

#ifdef TEST_CORRECTNESS

void correctness_tests ()
	/* try to obtain anomalous access */
{
	int n, count;

	CORRECTNESS_TEST (SV->s, S_lock, "S_lock", 
		{	lptr->lock(); 
			lock_body; 
			lptr->unlock();
		});
	CORRECTNESS_TEST (SV->t, T_lock, "T_lock", 
		{	short my_place = lptr->lock();
			lock_body;
			lptr->unlock(my_place);
		});
	CORRECTNESS_TEST (SV->a, A_lock, "A_lock", 
		{	short my_place = lptr->lock();
			lock_body;
			lptr->unlock(my_place);
		});
	CORRECTNESS_TEST (SV->m, M_lock, "M_lock", 
		{	lptr->lock(); 
			lock_body;
			lptr->unlock();
		});
}
#endif

#define A(s) asm (s);
void null_loop_test()
{
    int j;

    A("| --null loop test--");

    RUN_TEST_ALL ("null", 0, , { ; });
}

void delay_test()
{
    int j;

    A("| --delay test--");

    RUN_TEST_ALL ("delay", 0, , { delay();});
}


void simple_test()
{
	int j;
	init_S ();
A ("	| -----------------------slock test------------------------------")
	for(j = 0; j < num_nodes; j+=2) {
		RUN_TEST_ALL ("slock", j, 
			 register slock *lptr = &SV->s,
			 {slock_lock(lptr); slock_unlock(lptr); });
	}
}

void simple_lin_test()
{
	int j;
	init_S ();
A ("	| -----------------------slocklin test------------------------------")
	for(j = 0; j < num_nodes; j+=2) {
		RUN_TEST_ALL ("slocklin", j,
			register slock *lptr = &SV->s,
			{slocklin_lock(lptr); slocklin_unlock(lptr); });
	}
}

void ticket_test()
{
	int j;
	init_T ();
A ("	| -----------------------tlock test------------------------------")
	for(j = 0; j < num_nodes; j+=2) {
		RUN_TEST_ALL ("tlock",j,
			register tlock *lptr = &SV->t,
			{tlock_unlock(lptr, tlock_lock(lptr));});
	}
}


void ticketdel_test()
{
	int j;
	init_T ();
A ("	| -----------------------tlockdel test------------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("tlockdel", j,
			register tlock *lptr = &SV->t,
			{tlockdel_unlock(lptr, tlockdel_lock(lptr));});
	}
}

void simple_exp_test()
{
	int j;
	init_S ();
A ("	| -----------------------slockexp test------------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("slockexp", j, 
			register slock *lptr = &SV->s,
			{slockexp_lock(lptr); slockexp_unlock(lptr); });
	}
}

void mcs_test()
{
	int j;
	init_M ();
A ("	| -----------------------mlock test------------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("mlock", j, 
			register mlock *lptr = &SV->m, 
			{mlock_lock(lptr); mlock_unlock(lptr);});
	}
}




void anderson_test()
{
	int j;
A ("	| -----------------------alock test------------------------------")
	init_A ();
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("alock", j, 
			register alock *lptr = &SV->a, 
			{alock_unlock(lptr,alock_lock(lptr));});
	}
}


void timing_tests ()
{
	mcs_test();
	ticketdel_test();
	anderson_test();
	simple_exp_test();
	simple_lin_test();
	simple_test();
	ticket_test();
}

void tlockdel_test1(j)
int j;
{
	init_T ();
A ("	| -----------------------tlockdel test------------------------------")
	RUN_TEST_ALL ("tlockdel", j, 
			register tlock *lptr = &SV->t, 
			{tlockdel_unlock(lptr, tlockdel_lock(lptr));});
}

void anderson_test1(j)
int j;
{
A ("	| -----------------------alock test------------------------------")
	init_A ();
		RUN_TEST_ALL ("alock", j, 
			register alock *lptr = &SV->a, 
			{alock_unlock(lptr,alock_lock(lptr));});
}

void synchtest ()			/* MAIN */
{
	timing_tests();
}
void anderson_delay()
{
	int j;
A ("	| -----------------------alock delay ------------------------------")
	init_A ();
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("alock-delay", j, 
			register alock *lptr = &SV->a,
			{
				short place = alock_lock(lptr); 
				delay();
				alock_unlock(lptr,place);
			});
	}
}

void mcs_delay()
{
	int j;
	init_M ();
A ("	| -----------------------mlock delay test---------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("mlock-delay", j, 
			register mlock *lptr = &SV->m, 
			{	
				mlock_lock(lptr); 
				delay(); 
				mlock_unlock(lptr);
			});
	}
}

void ticketdel_delay()
{
	int j;
	init_T ();
A ("	| -----------------------tlockdel delay test-------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("tlockdel-delay", j, 
			register tlock *lptr = &SV->t, 
			{
				short ticket;
				ticket = (tlockdel_lock(lptr));
				delay();
				tlockdel_unlock(lptr,ticket);
			});
	}
}

void simple_exp_delay()
{
	int j;
	init_S ();
A ("	| -----------------------slockexp delay test-------------------------")
	for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("slockexp-delay", j, 
			register slock *lptr = &SV->s, 
			{
				slockexp_lock(lptr); 
				delay(); 
				slockexp_unlock(lptr); 
			});
	}
}

delay_timing_tests ()
{
	null_loop_test();
	delay_test();
	mcs_delay();
	anderson_delay();
	ticketdel_delay();
	simple_exp_delay();
}
