extern short vpid;
extern short num_nodes;
extern short nodes_in_test;
extern long repeat_count;

typedef unsigned long ticks;
static inline unsigned long ticks_to_ms (ticks t) { return t/16; }
static inline unsigned long ticks_to_us (ticks t) { return (t * 125) >> 1; }
static inline ticks ms_to_ticks (unsigned long m) { return m*16; }

#define REPEAT_COUNT 100000

extern ticks start_time, end_time; /* for use in following macro */
extern struct SV_struct *SV;

#define RUN_TEST(HIGHEST_NODE,DECLS,LOOP_BODY) \
		tight_barrier(); \
		if (HIGHEST_NODE >= vpid) {\
			short kmode = entkerin; \
			register int i; \
			DECLS; \
			asm volatile(".even");\
			start_time = rtc; \
			for (i=repeat_count; i--;) LOOP_BODY; \
			end_time = rtc; \
			AsAr = kmode; \
			SV->time[vpid] = end_time - start_time; \
			tight_barrier(); \
		} \
		else { SV->time[vpid] = 0; tight_barrier(); }

#define RUN_TEST_ALL(NAME,HIGHEST_NODE,DECLS,LOOP_BODY) {\
	int nodes_in_test, repeat_count;\
	nodes_in_test = HIGHEST_NODE + 1; \
	repeat_count = REPEAT_COUNT/nodes_in_test;\
	if (vpid == 0) { \
		printf("%s\n",NAME);\
	} \
	RUN_TEST (HIGHEST_NODE, DECLS, LOOP_BODY); \
	if (vpid == 0) { \
		int i = 0; \
		short pmax = 0; \
		ticks max = SV->time[0]; \
		for(i = 1; i <= HIGHEST_NODE; i++) \
			if (max < SV->time[i]) { max = SV->time[i]; pmax = i; } \
		printf("Spin Lock test %2d processors, %8d iterations/processor, time = %10d usec\n", nodes_in_test,repeat_count, ticks_to_us(max));\
	} \
	tight_barrier();\
}

#define CORRECTNESS_TEST(slock,type,NAME,lock_body) \
	{ \
		count = 0; \
		RUN_TEST_ALL (slock,type,NAME, (num_nodes - 1), lock_body, \
		{ \
			if (SV->scratch != pnn) count++; \
			SV->scratch = pnn; \
			if ((n = SV->scratch) != pnn) { \
				printf ("%x: failed; read %x, iteration %d\n", pnn, n, i); \
				i=2*REPEAT_COUNT; /* break loop */ \
			} \
		} \
		); \
		printf ("%x: %s, %d interleavings\n", pnn, NAME, count); \
	} 
