/***************************************************************** 
	slightly smarter: take ticket and spin on remote read
	guarantees FIFO ordering
	with coherent caches, this would spin locally until an unlock,
	at which point there would be a flurry of bus traffic and
	then everybody would settle in again
 *****************************************************************/

#include <public.h>

#include "ticket.h"

#ifndef USE_ASM

	fail
#endif USE_ASM

#define A(s) asm volatile(s);

static struct PNC_am ticket_cta_block;

void init_T ()
{
	/* the hand-coded assembler uses its own special ctl blk for speed */
	ticket_cta_block.am_mask = 0;
	ticket_cta_block.am_incr = 1;
}



static void ticket()
{
A ("	| assembly version of ticket lock 									")
A ("	| parameter: pointer to lock in a1									")
A ("	.text																")
A ("	.even 																")
A (".globl tlock_lock														")
A ("tlock_lock:																")
A ("	| don't need an fp													")
A ("	| no non-volatiles to save											")
A ("	movl #_ticket_cta_block,a0											")
A ("	| fill in control block for atomic increment 						")
A ("	movl a1,a0@				| fill in &tlock.next_ticket in ctl blk		")
A ("	movl a0,0xffffd100		| fetch-and-phi --> my_ticket 				")
A ("	movw a0@(8),d0			| put my_ticket in d0						")
A ("	cmpw a1@(2),d0			| is my ticket currently being served?		")
A ("	bne tdo_poll			| continue polling							")
A ("	rts																	")
A ("tdo_poll:																")
A ("	movl a7@+,a0			| pop return address (setup fast return)	")
A ("tpoll_again:															")
A ("	cmpw a1@(2),d0			| while(now_serving != my_ticket)			")
A ("	bne tpoll_again			| if not, poll again						")
A ("	jmp a0@					| return to caller 							")
}

static void ticketdel()
{
A ("	| assembly version of ticket lock 									")
A ("	| parameter: pointer to lock in a1									")
A ("	.text																")
A ("	.even 																")
A (".globl tlockdel_lock													")
A ("tlockdel_lock:															")
A ("	| don't need an fp													")
A ("	| no non-volatiles to save											")
A ("	movl #_ticket_cta_block,a0											")
A ("	| fill in control block for atomic increment 						")
A ("	movl a1,a0@				| fill in &tlock.next_ticket in ctl blk		")
A ("	movl a0,0xffffd100		| fetch-and-phi --> my_ticket 		 		")
A ("	movw a0@(8),d0			| put my_ticket in d0						")
A ("	cmpw a1@(2),d0			| is my ticket currently being served?		")
A ("	bne tdel_poll_setup		| not equal --> poll with prop. delay		")
A ("	rts																	")
A ("tdel_poll_setup:														")
A ("	movl a7@+,a0			| pop return address (setup fast return)	")
A ("	bra tdel_do_poll		| poll with proportional delay				")
A ("tdel_delay:																")
A ("	subw #2,d1				| subtract off one							")
A ("	blt tdel_do_poll		| poll with proportional delay				")
A ("tdel_delay_loop:														")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	nop																	")
A ("	dbra d1,tdel_delay_loop	| iterations = (diff -1) 	 				")
A ("tdel_do_poll:															")
A ("	movw d0,d1				| my_ticket 								")
A ("	subw a1@(2),d1			| my_ticket - now_serving 					")
A ("	bne tdel_delay			| not equal --> poll with prop. delay		")
A ("	jmp a0@					| return to caller 							")
}


static void ticketdelinc()
{
A ("	| assembly version of ticket lock 									")
A ("	| parameter: pointer to lock in a1									")
A ("	.text																")
A ("	.even 																")
A (".globl tlockdelinc_lock													")
A ("tlockdelinc_lock:														")
A ("	| don't need an fp													")
A ("	| no non-volatiles to save											")
A ("	movl #_ticket_cta_block,a0											")
A ("	| fill in control block for atomic increment 						")
A ("	movl a1,a0@				| fill in &tlock.next_ticket in ctl blk		")
A ("	movl a0,0xffffd100		| fetch-and-phi --> my_ticket 		 		")
A ("	movw a0@(8),d0			| put my_ticket in d0						")
A ("	cmpw a1@(2),d0			| is my ticket currently being served?		")
A ("	bne tdelinc_poll_setup		| not equal --> poll with prop. delay	")
A ("	rts																	")
A ("tdelinc_poll_setup:														")
A ("	movw #-1,a0				| initialize last delay interval count 		")
A ("	movl d2,a7@-			| free up a register						")
A ("	bra tdelinc_do_poll		| poll with proportional delay				")
A ("tdelinc_delay:															")
A ("	movw a0,d2									")
A ("	blt tdelinc_no_old_diff	| poll with proportional delay				")
A ("	subw d1,d2															")
A ("tdelinc_no_old_diff:													")
A ("	movw d1,a0				| save original difference					")
A ("	subw #1,d1				| subtract off one							")
A ("	beq tdelinc_do_poll		| ticket diff too small, don't delay		")
A ("	aslw #4,d1				| compute delay iterations 15 * (diff - 1)	")
A ("	subw a0,d1				| ...										")
A ("tdelinc_delay_loop:														")
A ("	dbra d1,tdelinc_delay_loop	| delay							 		")
A ("tdelinc_do_poll:														")
A ("	movw d0,d1				| my_ticket 								")
A ("	subw a1@(2),d1			| my_ticket - now_serving 					")
A ("	bne tdelinc_delay			| not equal --> poll with prop. delay	")
A ("	movl a7@+,d2			| restore register							")
A ("	rts																	")
}
