#define NULL	0
#define P 		18
#define LogP	5

typedef unsigned char boolean;
#define False ((boolean) 0)
#define True ((boolean) 1)

typedef unsigned char role_enum;
#define LOSER ((role_enum) 1)
#define WINNER ((role_enum) 2)
#define CHAMPION ((role_enum) 4)
#define NOOP ((role_enum) 8)
#define DROPOUT ((role_enum) 16)

typedef struct {
	boolean *opponent;
	role_enum role;
	boolean flag;
} round_t;


shared round_t rounds[P][LogP];
shared boolean champion_sense = False;

private boolean sense = True;
private round_t *myrounds;

extern private int vpid;
extern shared int num_nodes;

void tournament_barrier_init()
{
	int k, log_num_nodes = 0;
	myrounds = &rounds[vpid][0];
	for(k = 1; k < num_nodes; log_num_nodes++, k <<= 1); 
	for(k = 0; k < LogP; k++) myrounds[k].role = DROPOUT; /* default */
	for(k = 0; k < log_num_nodes; k++) {
		int vpid_mod_2_sup_k_plus_1 = vpid % (1 << (k+1));
		if (vpid_mod_2_sup_k_plus_1 == 0) {
			int partner = vpid + (1 << k);
			if (partner < num_nodes) myrounds[k].role = WINNER;
			else myrounds[k].role = NOOP;
			myrounds[k].opponent = (boolean *) NULL;
		}
		else if (vpid_mod_2_sup_k_plus_1 == (1 << k)) {
			myrounds[k].role = LOSER;
			myrounds[k].opponent = &rounds[vpid - (1<<k)][k].flag;
			break;
		}
	}
	if (vpid == 0) myrounds[k-1].role = CHAMPION;
	for(k = 0; k < LogP; k++) myrounds[k].flag = False; 
}

void tournament_barrier()
{
	register round_t *round = myrounds;
	for(;;) {
		if(round->role & LOSER) {
			*(round->opponent) = sense;
			while (champion_sense != sense);
			break;
		} else if (round->role & WINNER) while (round->flag != sense);
		else if (round->role & CHAMPION) {
			while (round->flag != sense);
			champion_sense = sense;
			break;
		}
		round++;
	}
	sense ^= True;
}
