#include <us.h>
#include "mcs.h"

MCSlock *LOCK;

int iterations = 1;
int nodes;
long *times;

usage()
{
	printf("usage: main [-i iterations]\n");
	exit(1);
}

void synchtest(); /* forward declaration */

main(argc,argv)
int argc;
char **argv;
{
    int i;

    while (--argc > 0) {
	switch((*++argv)[1]) {
	case 'i':
	    if (--argc < 0) usage();
	    iterations = atoi(*++argv);
	    break;
	default:
	    usage();
	}
    }

    InitializeUs();

    nodes = TotalProcsAvailable();
    ShareBlk(&nodes,sizeof(nodes));

    if (nodes > 1) dissemination_barrier_init(nodes);

    LOCK = (MCSlock*) UsAlloc(sizeof(MCSlock));
    *LOCK = 0;
    ShareBlk(&LOCK,sizeof(MCSlock*));

    times = (long*) UsAlloc(nodes * sizeof(long));
    ShareBlk(&times,sizeof(long*));

    GenTaskForEachProc(synchtest,iterations/nodes);
    return 0;
}

void synchtest(iter)
int iter;
{
    long start,end,maxtime;
    int i;
    qnode *mylink = (qnode *) UsAllocLocal(sizeof(qnode));

    if (nodes > 1) dissemination_barrier();

    start = getusecclock();
    for(i=iter;i>0;i--) {
	    mcs_lock(LOCK,mylink);
#ifdef DEBUG
	    printf("%d has lock\n",UsProc_Node);
	    sleep(1);
	    printf("%d releasing lock\n",UsProc_Node);
#endif
	    mcs_unlock(LOCK,mylink);
    }
    end = getusecclock();
    times[UsProc_Node] = end - start;

    if (nodes > 1) dissemination_barrier();

    for(maxtime = times[0],i=1;i<nodes;i++) 
	    if(times[i] > maxtime) maxtime = times[i];

    if (UsProc_Node == 0)
	printf("%d processors, %d iterations (%d each): %d usec elapsed\n",
		nodes, iterations, iter, maxtime);
}
